
/*
  Released under the antivirial license.  Basically, you can do anything
  you want with it as long as what you want doesn't involve the GNU GPL.
  See http://www.ecstaticlyrics.com/antiviral/ for more information.
*/

#ifdef WINDOWS
  #define __MSVCRT_VERSION__ 0x0601
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#ifdef WINDOWS
  #include "glfw.h"
  #define S_ISLNK(x) 0
#else
  #include <GL/glfw.h>
#endif

#include <GL/glu.h>
#include "stb_image.h"
#include "sha1.h"

#define MARGIN 1

struct int_xy {
  int x;
  int y;
};

struct item { // A file or a folder...
  char *name;
  char *pathname;
  long long size;
  struct int_xy one;
  struct int_xy two;
  struct item *parent;
  int is_a_folder;
  int index;
  struct item **contents;
  float r, g, b;
};

int filesystem;
struct item *current_item;
int window_width;
int window_height;

GLint font_texture = 0;

GLuint load_image(char *filename) {
  int width, height;
  char *image = stbi_load(filename, &width, &height, NULL, 4);
  if (image == NULL) {
    fprintf(stderr, "Failed to load %s\n", filename);
    exit(1);
  };
  GLuint texture;
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
  stbi_image_free(image);
  return texture;
};

static void sort_items(struct item **array, int count) {
  if (count <= 1) return;
  int half_one = count >> 1;
  int half_two = count - half_one;
  sort_items(array, half_one);
  sort_items(array + half_one, half_two);
  struct item **temp = malloc(sizeof(void *) * count);
  struct item **pointer_one = array;
  struct item **pointer_two = array + half_one;
  for (int i = 0; i < count; i++) {
    if (half_one && half_two) {
      if ((*pointer_one)->size >= (*pointer_two)->size) {
        temp[i] = *pointer_one;
        pointer_one++;
        half_one--;
      } else {
        temp[i] = *pointer_two;
        pointer_two++;
        half_two--;
      };
    } else if (half_one) {
      temp[i] = *pointer_one;
      pointer_one++;
      half_one--;
    } else {
      temp[i] = *pointer_two;
      pointer_two++;
      half_two--;
    };
  };
  memmove(array, temp, sizeof(void *) * count);
  free(temp);
};

int mouse_x = 0;
int mouse_y = 0;

void GLFWCALL position_callback(int x, int y) {
  mouse_x = x;
  mouse_y = window_height - y - 1;
};

int window_refresh_flag = 1;
void GLFWCALL window_refresh_callback() {
  window_refresh_flag = 1;
};

struct item *what_was_clicked_on(int allow_nested_clicks) {
  struct item *check = current_item;
  REPEAT:
  for (int i = 0; i < check->index; i++) {
    if (check->contents[i]->one.x <= mouse_x && mouse_x <= check->contents[i]->two.x && check->contents[i]->one.y <= mouse_y && mouse_y <= check->contents[i]->two.y) {
      check = check->contents[i];
      if (allow_nested_clicks) {
        goto REPEAT;
      } else {
        break;
      };
    };
  };
  printf("Click Location: %s\n", check->pathname);
  if (check == current_item) return NULL;
  return check;
};

#ifndef WINDOWS

char *escape(char *text) {
  // Escape text strings for a Perl script...
  char *new = malloc(4 * strlen(text) + 1);
  char *j = new;
  for (int i = 0; text[i] != 0; i++) {
    if (text[i] == '\\') {
      *j++ = '\\'; *j++ = '\\';
    } else if (text[i] == '"') {
      *j++ = '\\'; *j++ = '"';
    } else if (text[i] == '$') {
      *j++ = '\\'; *j++ = '$';
    } else if (text[i] == '@') {
      *j++ = '\\'; *j++ = '@';
    } else if (text[i] < 32 || text[i] > 126) {
      int a = (text[i] >> 4) + 48; if (a >= 58) a += 7;
      int b = (text[i] & 15) + 48; if (b >= 58) b += 7;
      *j++ = '\\'; *j++ = 'x'; *j++ = a; *j++ = b;
    } else {
      *j++ = text[i];
    };
  };
  *j++ = 0;
  return realloc(new, strlen(new) + 1);
};

FILE *script = NULL;

void delete(struct item *it) {

  static int nest = 0;

  // This delete function does not delete anything!

  char *escaped = escape(it->pathname);

  if (it->is_a_folder) {
    for (int i = 0; i < it->index; i++) {
      nest++;
      delete(it->contents[i]);
      nest--;
    };
    fprintf(script, "    [%d, \"%s\"],\n", 1 + (!!nest << 1), escaped);
  } else {
    fprintf(script, "    [%d, \"%s\"],\n", 0 + (!!nest << 1), escaped);
  };

  free(escaped);

};

#endif

void GLFWCALL button_callback(int button, int action) {
  if (action != GLFW_PRESS) return;
  window_refresh_flag = 1;
  if (button == 0 || button == 1) {
    struct item *selected_item = what_was_clicked_on(1);
    if (selected_item != NULL && selected_item->parent != NULL) {
      #ifndef WINDOWS
      if (button == 0) delete(selected_item);
      #endif
      long long size = selected_item->size;
      struct item *parent = selected_item->parent;
      for (int i = 0; i < parent->index; i++) {
        if (parent->contents[i] == selected_item) {
          memmove(parent->contents + i, parent->contents + i + 1, (parent->index - i - 1) * sizeof(void *));
          parent->index--;
          break;
        };
      };
      while (1) {
        parent->size -= size;
        parent = parent->parent;
        if (parent == NULL) break;
        sort_items(parent->contents, parent->index);
      };
    };
  } else if (button == 2) {
    struct item *selected_item = what_was_clicked_on(1);
    if (selected_item != NULL) current_item = selected_item;
  };
};

void GLFWCALL wheel_callback(int position) {
  window_refresh_flag = 1;
  static int memory = 0;
  for (; memory < position; memory++) {
    struct item *selected_item = what_was_clicked_on(0);
    if (selected_item != NULL) {
      current_item = selected_item;
    };
  };
  for (; memory > position; memory--) {
    if (current_item->parent != NULL) {
      current_item = current_item->parent;
    };
  };
};

int exit_flag = 0;

void GLFWCALL key_callback(int key, int action) {
  if (action != GLFW_PRESS) return;
  if (key == GLFW_KEY_ESC) exit_flag = 1;
};

void scan_folder(struct item *folder) {
  #define folder (*folder)

//  printf("Scanning folder '%s'\n", folder.pathname);

  folder.is_a_folder = 1;

  DIR *dir;
  dir = opendir(folder.pathname);
  if (dir == NULL) {
    fprintf(stderr, "Error %d while opening '%s'\n", errno, folder.pathname);
  };
  struct dirent *entry;
  while (entry = readdir(dir)) {
    if (entry->d_name[0] == '.') {
      if (entry->d_name[1] == 0) continue;
      if (entry->d_name[1] == '.') {
        if (entry->d_name[2] == 0) continue;
      };
    };
    folder.contents = realloc(folder.contents, (folder.index + 1) * sizeof(void *));
    folder.contents[folder.index] = malloc(sizeof(struct item));
    memset(folder.contents[folder.index], 0, sizeof(struct item));
    #define that (*folder.contents[folder.index])
    int length = strlen(entry->d_name);
    that.name = malloc(length + 1);
    strcpy(that.name, entry->d_name);
    if (folder.pathname[strlen(folder.pathname) - 1] == '/') {
      that.pathname = malloc(strlen(folder.pathname) + length + 1);
      sprintf(that.pathname, "%s%s", folder.pathname, that.name);
    } else {
      that.pathname = malloc(strlen(folder.pathname) + 1 + length + 1);
      sprintf(that.pathname, "%s/%s", folder.pathname, that.name);
    };
    that.parent = &folder;
    #ifdef WINDOWS
    struct __stat64 info;
    if (_stat64(that.pathname, &info) == -1) {
    #else
    struct stat info;
    if (lstat(that.pathname, &info) == -1) {
    #endif
      fprintf(stderr, "Error %d while stat()ing '%s'\n", errno, that.pathname);
    } else {
      if (info.st_dev == filesystem) {
        if (S_ISLNK(info.st_mode)) {
        };
        if (S_ISDIR(info.st_mode)) {
          int length = strlen(that.pathname);
          that.pathname = realloc(that.pathname, length + 2);
          that.pathname[length] = '/';
          that.pathname[length + 1] = 0;
          scan_folder(&that);
        };
        if (S_ISREG(info.st_mode)) {
          struct item *current = &that;
          while (current != NULL) {
            current->size += info.st_size;
            current = current->parent;
          };
        };
      };

      // Assign a color based on hash of the pathname...
      char sum[20];
      sha1(sum, that.pathname, strlen(that.pathname));
      that.r = 0.6 * sum[0] / 255.0 + 0.2;
      that.g = 0.6 * sum[1] / 255.0 + 0.2;
      that.b = 0.6 * sum[2] / 255.0 + 0.2;

      #undef that
      folder.index++;
    };
  };
  closedir(dir);
  sort_items(folder.contents, folder.index);
  #undef folder
};

void draw_text(int x, int y, char *text) {
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND); glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glBindTexture(GL_TEXTURE_2D, font_texture);
  glBegin(GL_QUADS);
  for (int i = 0; i < strlen(text); i++) {
    int c = text[i];
    if (c < 32 || c > 126) c = 127;
    int cx = c % 16;
    int cy = c / 16;
    glTexCoord2d((cx + 0) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x, y);
    glTexCoord2d((cx + 1) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x + 8, y);
    glTexCoord2d((cx + 1) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x + 8, y + 16);
    glTexCoord2d((cx + 0) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x, y + 16);
  };
  glEnd();
  glDisable(GL_TEXTURE_2D);
};

void draw_folder (struct item *folder, int nest, int index) {
  #define folder (*folder)
  struct int_xy one = folder.one;
  struct int_xy two = folder.two;

  int width = two.x - one.x;
  int height = two.y - one.y;

  one.x += MARGIN; two.x -= MARGIN; width -= 2 * MARGIN;
  one.y += MARGIN; two.y -= MARGIN; height -= 2 * MARGIN;

  if (width < 1 || height < 1) return;

  if (nest == 0) {
    glColor3f(0.0, 0.0, 0.0);
  } else {
    struct item *source = &folder;
    for (int i = 1; i < nest; i++) source = source->parent;
    float r = source->r;
    float g = source->g;
    float b = source->b;
    for (int i = 1; i < nest; i++) {
      r *= 0.8; g *= 0.8; b *= 0.8;
    };
    glColor3f(r, g, b);
  };

  glBegin(GL_QUADS);
  glVertex2f(one.x, one.y);
  glVertex2f(two.x, one.y);
  glVertex2f(two.x, two.y);
  glVertex2f(one.x, two.y);
  glEnd();

  one.x += MARGIN; one.y += MARGIN; width -= 2 * MARGIN;
  two.x -= MARGIN; two.y -= MARGIN; height -= 2 * MARGIN;

  if (width >= 16 && height >= 16) {

    char *name = (nest ? folder.name: folder.pathname);
    int name_len = strlen(name);

    char size[16];
    if (folder.size > 1000000000) {
      snprintf(size, 16, "%.3g GB", folder.size / 1000000000.0);
    } else if (folder.size > 1000000) {
      snprintf(size, 16, "%.3g MB", folder.size / 1000000.0);
    } else if (folder.size > 1000) {
      snprintf(size, 16, "%.3g kB", folder.size / 1000.0);
    } else {
      snprintf(size, 16, "%.3g B", folder.size / 1.0);
    };
    size[15] = 0;
    int size_len = strlen(size);

    if (nest < 2) {
      glColor3f(1.0, 1.0, 1.0);
    } else {
      glColor3f(0.0, 0.0, 0.0);
    };

    int max = (width - 4) / 8;

    if (width >= height || name_len <= max) {
      if (name_len <= max) {
        draw_text(one.x + 2, two.y - 16 - 1, name);
      } else {
        char temp[max + 1];
        memmove(temp, name, max);
        temp[max] = 0;
        temp[max-1] = 127;
        draw_text(one.x + 2, two.y - 16 - 1, temp);
      };
      if (two.x - one.x > 8 * (name_len + size_len + 2) + 4) {
        draw_text(two.x - size_len * 8 - 2, two.y - 16 - 1, size);
      };
      two.y -= 16 + MARGIN;
      height -= 16 + MARGIN;
    } else {
      glPushMatrix();
      glTranslated(one.x + 16 + 1, one.y + 2, 0);
      glRotated(90, 0, 0, 1);
      max = (height - 4) / 8;
      if (name_len <= max) {
        draw_text(0, 0, name);
      } else {
        char temp[max + 1];
        memmove(temp, name, max);
        temp[max] = 0;
        temp[max-1] = 127;
        draw_text(0, 0, temp);
      };
      if (two.y - one.y > 8 * (name_len + size_len + 2) + 4) {
        draw_text(two.y - one.y - size_len * 8 - 4, 0, size);
      };
      glPopMatrix();
      one.x += 16 + MARGIN;
      width -= 16 + MARGIN;
    };

  };

  int first = 0;
  int consumed_y = 0;
  int remaining_y = height;
  long long remaining_size = folder.size;

  // Determine the best number of items to show on each row within the
  // folder such that the ratio between width and height is most equal.

  // We do this by first assuming we'll put every item in a single row,
  // then we pop items off the end until the situation ceases to improve.

  while (first < folder.index) {

    if (width < 2 * MARGIN + 1) return;
    if (height - consumed_y < 2 * MARGIN + 1) return;

    int last = folder.index - 1;
    long long row_size = remaining_size;

    float previous_worst_ratio = 0.0;

    long long selected_row_size = row_size;
    int selected_last = last;

    while (1) {
      float estimated_y = (float) remaining_y * row_size / remaining_size;
      long long first_size = folder.contents[first]->size;
      long long last_size = folder.contents[last]->size;
      float first_width = (float) width * first_size / row_size;
      float last_width = (float) width * last_size / row_size;
      float first_ratio, last_ratio;

      if (estimated_y >= 16 + 2 * MARGIN + 2) estimated_y *= 1.0f;

      if (first_width < estimated_y) {
        first_ratio = (float) first_width / estimated_y;
      } else {
        first_ratio = (float) estimated_y / first_width;
      };
      if (last_width < estimated_y) {
        last_ratio = (float) last_width / estimated_y;
      } else {
        last_ratio = (float) estimated_y / last_width;
      };
      float worst_ratio;
      if (first_ratio < last_ratio) {
        worst_ratio = first_ratio;
      } else {
        worst_ratio = last_ratio;
      };
      if (worst_ratio < previous_worst_ratio) break;
      previous_worst_ratio = worst_ratio;
      selected_row_size = row_size;
      selected_last = last;
      if (first == last) break;
      row_size -= folder.contents[last]->size;
      last--;
    };

    row_size = selected_row_size;
    last = selected_last;

    int actual_y = ceilf((float) remaining_y * row_size / remaining_size);

    if (actual_y < 2 * MARGIN + 1) break;

    long long sum = 0;
    for (int i = first; i <= last; i++) {
      int sx = roundf((float) width * sum / row_size);
      sum += folder.contents[i]->size;
      int tx = roundf((float) width * sum / row_size);
      folder.contents[i]->one.x = one.x + sx;
      folder.contents[i]->two.x = one.x + tx;
      //folder.contents[i]->one.y = one.y + consumed_y;
      //folder.contents[i]->two.y = one.y + consumed_y + actual_y;
      folder.contents[i]->two.y = two.y - consumed_y;
      folder.contents[i]->one.y = two.y - consumed_y - actual_y;
//      if (nest < 2) printf("Draw '%s' at %d, %d, %d, %d\n", folder.contents[i]->name, folder.contents[i]->one.x, folder.contents[i]->two.x, folder.contents[i]->one.y, folder.contents[i]->two.y);
      draw_folder(folder.contents[i], nest + 1, index == -1 ? i : index);
    };

    remaining_size -= row_size;
    consumed_y += actual_y;
    remaining_y -= actual_y;
    first = last + 1;

  };

  #undef folder
};

int main(int argc, char **argv) {

  struct item root = {};
  root.parent = NULL;

  if (argc >= 2) {
    root.name = argv[1];
  } else {
    root.name = "/";
  };

  root.pathname = root.name;

  #ifdef WINDOWS
  struct __stat64 info;
  if (_stat64(root.pathname, &info) == -1) {
  #else
  struct stat info;
  if (lstat(root.pathname, &info) == -1) {
  #endif
    fprintf(stderr, "Error %d while stat()ing '%s'\n", errno, root.pathname);
    exit(1);
  };
  filesystem = info.st_dev;

  scan_folder(&root);

  current_item = &root;

  glfwInit();
  glfwOpenWindow(896, 672, 8, 8, 8, 0, 0, 0, GLFW_WINDOW);
  glfwSetWindowTitle("Moosfet's File Size Viewer");
  glfwSetMousePosCallback(position_callback);
  glfwSetMouseButtonCallback(button_callback);
  glfwSetMouseWheelCallback(wheel_callback);
  glfwSetKeyCallback(key_callback);
  glfwSetWindowRefreshCallback(window_refresh_callback);

  {
    extern char data_font;
    extern int size_font;
    int width, height;
    char *image = stbi_load_from_memory(&data_font, size_font, &width, &height, NULL, 4);
    if (image == NULL) {
      fprintf(stderr, "Failed to load font from memory!\n");
      exit(1);
    };
    GLuint texture;
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
    stbi_image_free(image);
    font_texture = texture;
  };

  #ifndef WINDOWS

  script = fopen("script.pl", "wb");
  if (script == NULL) {
    fprintf(stderr, "Failed to create a new file named 'script.pl'\n");
    exit(1);
  };

  fprintf(script, "#!/usr/bin/perl\n");
  fprintf(script, "require \"functions.pl\";\n\n");
  fprintf(script, "# Uncomment the line for the operation you wish to perform!\n\n");

  fprintf(script, "#copy_the_files_to(\"/home/temporary/\");\n\n");
  fprintf(script, "#move_the_files_to(\"/home/temporary/\");\n\n");
  fprintf(script, "#delete_the_files();\n\n");
  fprintf(script, "#do_something_else(); # Make sure you know what that is!\n\n");

  fprintf(script, "# The rest of this file is just the list of files you selected.\n\n");
  fprintf(script, "sub the_list_of_files {\n");
  fprintf(script, "  @list = (\n");

  #endif

  while (glfwGetWindowParam(GLFW_OPENED) && !exit_flag) {

    if (window_refresh_flag) {

      glfwGetWindowSize(&window_width, &window_height);
      glViewport(0, 0, window_width, window_height);
      glMatrixMode(GL_MODELVIEW); glLoadIdentity();
      glTranslated(-1.0, -1.0, 0.0);
      glScaled(2.0 / window_width, 2.0 / window_height, 1.0);

      current_item->one.x = 0;
      current_item->one.y = 0; // 16 + MARGIN;
      current_item->two.x = window_width;
      current_item->two.y = window_height;

      static struct item *last_selected = NULL;
      //selected_item = NULL;

      glClearColor(0.0, 0.0, 0.0, 1.0);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

      draw_folder(current_item, 0, -1);

      #if 0
      glColor3f(1.0, 1.0, 1.0);
      if (selected_item) draw_text(MARGIN + 3, MARGIN, selected_item->pathname);

      if (selected_item != last_selected) {
        if (selected_item) {
          //glfwSetWindowTitle(selected_item->pathname);
        } else {
          //glfwSetWindowTitle("Moosfet's File Size Viewer");
        };
        last_selected = selected_item;
      };
      #endif

      printf("Redrawing the screen!\n");
      glfwSwapBuffers();

      window_refresh_flag = 0;
    };

    glfwWaitEvents();

  };

  #ifndef WINDOWS

  fprintf(script, "  );\n");
  fprintf(script, "};\n");
  fclose(script);

  #endif

  glfwCloseWindow();

};
